/***************************************************************************
 *   Copyright (C) 2006 by M.Fedotov   *
 *   anyremote@mail.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/rfcomm.h>

#include <kuniqueapplication.h>

#include "bt_comm.h"
#include "ar_parser.h"
#include "ar_client.h"
#include "logger.h"

extern AR_Parser*   	    conf;
extern AR_Client*   	    theClient;
extern logger2*     	    theLog;
extern KUniqueApplication*  pApp;

BtComm::BtComm()
{
	portfd       = 0;
	sportfd      = 0;
	portflags    = 0;
	oldflags     = 0;
	rtscts       = 0;

	style	     = SERVER_NO;

	Connected    = false;
	useQuestionInCLCC = 0;
}

BtComm::~BtComm()
{
}

bool BtComm::isConnected()
{
	return Connected;
}

void BtComm::setConnected(bool c)
{
	Connected = c;
}

// Send AT+CKPD command
int BtComm::sendCKPD(const QString& key)
{   
	QString answer;

	theLog->debug("BtComm::sendCKPD() >" + key + "<");  

	if (key.isNull()) {
		theLog->debug("Empty symbol in sendCKPD()");
		return EXIT_NOK;
	}

	if (conf->getModel() == MODEL_SE || conf->getModel() == MODEL_SIEMENS) {    // SE-K700, Siemens-S55
		// If we send AT+CKPD=X
		// SE phone will replay
		// OK
		// +CKEV: X,1
		// +CKEV: X,0
		// And we need to filter out that CKEVs

		QString expect = QString(DEF_CKEV) + " " + key + ",0";
		sendCommand(QString(DEF_AT_CKPD) + QString("\"") + key + QString("\"\r"), answer, expect);
	
		theLog->debug("Expect " + expect);
		theLog->debug("Got all");

		// Filter out button up
		answer.remove(expect);
		expect = QString(DEF_CKEV) + " " + key + ",1";

		// Filter out button down
		answer.remove(expect);

		theLog->debug("Filtered " + answer);
	} else {
		sendCommand(QString(DEF_AT_CKPD) + QString("\"") + key + QString("\"\r"), answer, QString::null);
	}

	return theClient->handleCommand((char *)answer.ascii());
}


int BtComm::sendSeq(const QString& seq) 
{
	theLog->debug("BtComm::sendSeq()");
	if (! conf->getUseScreen()) {
		return EXIT_OK;
	}

	if (seq.isNull() || seq.isEmpty()) {
		theLog->debug("sendSeq() input is NULL");
		return EXIT_OK;
	}

	theLog->info("sendSeq() >" + seq + "<");

	int i = 0;
	if (theClient->serverMode > 0) {
		theLog->error("Can not use SendCKPD() in Server style");
		// Separate them
		//writeSocketConn(seq);
	} else {
		
		while (true) {
			QString ckpd = seq.section(' ', i, i);

			if (ckpd.isNull()) {
				theLog->debug("sendSeq() end ...");
				break;
			}
			if (ckpd == QString(PAUSE_STR)) {
				// Just wait a second
				theLog->debug("sendSeq() PAUSE");
				sleep(WAIT_TIME);
			} else {
				int ret = sendCKPD(ckpd);
				if (ret == EXIT_DISCON || ret == EXIT_ABORT) {
					theLog->info("Got disconnect or cancel or abort...");
					return ret;
				}
			}

			i++;
		}
	}
	return EXIT_OK;
}

// This is only for AT-style
bool BtComm::initPort(QString& port) 
{
	theLog->debug("BtComm::initPort() " + port);

	style = SERVER_NO;

	// get a file descriptor
	if ((portfd = open(port.ascii(), O_RDWR | O_NOCTTY | O_NDELAY)) < 0) {
		theLog->error(QString("can't open ") + port);
		portfd = 0;
		return false;
	}
	Connected = true;

	pApp->processEvents();

	// check to see that the file is a terminal
	if (!isatty(portfd)) {
		theLog->error(port + QString(" does not refer to a terminal"));
		return false;
	}
	pApp->processEvents();

	// get port attributes, store in oldterm
	if (tcgetattr(portfd, &oldterm) < 0) {
		theLog->error(QString("Can't get port attributes (") + port + ")");
		return false;
	}
	pApp->processEvents();

	// get port flags, save in oldflags
	if ((oldflags = fcntl(portfd, F_GETFL)) < 0) {
		theLog->error(QString("Can't get port flags (") + port + ")");
		return false;
	}
	pApp->processEvents();

	portterm = oldterm;
	portflags = oldflags;

	portterm.c_cflag = conf->getBaudrate() | CRTSCTS | CS8 | CLOCAL | CREAD | O_NDELAY;
	if (!rtscts) {
		portterm.c_cflag &= ~CRTSCTS;
	}
	portterm.c_iflag	= IGNPAR;
	portterm.c_oflag	= 0;
	portterm.c_lflag	= 0;
	portterm.c_cc[VTIME]	= 0;
	portterm.c_cc[VMIN]	= 0;

	tcflush(portfd, TCIOFLUSH);

	pApp->processEvents();

	if (tcsetattr(portfd,TCSANOW,&portterm) < 0) {
		theLog->error(QString("Can't set port attributes (") + port + ")");
		return false;
	}
	pApp->processEvents();

	// set non-blocking
	if (fcntl(portfd, F_SETFL, (portflags |= O_NONBLOCK)) < 0) {
		theLog->error(QString("Can't set port flags (") + port + ")");
		return false;
	}

	return true;
}

// Then port will be closed from a forked child use 0
int BtComm::closePort(int final) 
{
	theLog->debug("BtComm::closePort()");
	int retval = 0;

	Connected = false;

	if (theClient->serverMode > 0) {    // Server style
		
		if (portfd) { 
			close(portfd);
			portfd = 0;
		}
		if (sportfd) { 
			close(sportfd);
			sportfd = 0;
		}
	} else {			    // AT style
		
		if (!portfd) {
			return -1;
		}
	
		if (final) {
			// restore old settings
			if (tcsetattr(portfd, TCSADRAIN, &oldterm) < 0) {
				retval = -1;
			}

			if (fcntl(portfd, F_SETFL, oldflags) < 0) {
				retval = -1;
			}
		}
		close(portfd);
		portfd = 0;
	}
	return retval;
}

int BtComm::readTmo(char* buf, int len, int timeout) 
{
	//theLog->info("BtComm::readTmo()");

	int ok;
	fd_set rdfds;
	struct timeval tv;

	FD_ZERO(&rdfds);
	FD_SET(portfd, &rdfds);

	// block until something to read or timeout occurs.  select() is damn cool
	if (timeout < 0) {
		ok = select(portfd + 1, &rdfds, NULL, NULL, NULL);
	} else {
		tv.tv_sec = timeout / 1000000;
		tv.tv_usec = (timeout % 1000000);
		ok = select(portfd + 1, &rdfds, NULL, NULL, &tv);
	}
	pApp->processEvents();

	if (ok > 0) {
		ok = read(portfd, buf, len);
		if (ok == 0) {
			theLog->info("EOF during read()");
			Connected = false;
			return EOF;
		}
		return ok;
	} else {
		//Connected = false;
		return 0;
		//if (ok < 0) {
		//  return -1;
		//} else {
		//  errno = ETIMEDOUT;
		//  return 0;
		//}
	}
}

int  BtComm::bt_readchar(char* c, int timeout) 
{
	char buf[1];
	*c = 0;
        
	int rv =  readTmo(buf, 1, timeout);

	if (rv == 1) {
        	*c =  buf[0];
        	return 1;
        }
	if (rv == 0) return EOF - 1;

	return EOF;
}

int BtComm::sendCommand(const QString cmd, QString& answer, QString expect)
{
	return putCommand(cmd, answer, READ_TIMEOUT, expect);
}

int BtComm::putCommand(QString command,
		   QString& Answ,
		   int   timeout,
		   QString expect) 
{
	theLog->debug("BtComm::putCommand()");

	int count=0;
	int readcount;
	QString Answer;
	char tmp2[100];
	int timeoutcounter=0;
	int found=0;

	theLog->debug(QString("W: ") + command);

	// send command
	if (!(command.isNull() || command.isEmpty())) {
		write(portfd, command.ascii() ,strlen(command.ascii()));
		tcdrain(portfd);
	}

	do {
		pApp->processEvents();

		// This is very old code. Initially it was a part of btxmms project.
		// It was written on plain C.

		// try to read some bytes.
		usleep(100000);
		timeoutcounter++;

		// read data
		readcount=read(portfd,tmp2,sizeof(tmp2)-1);
		if (readcount<0) {
			readcount=0;
		}
		tmp2[readcount]=0;
		   
		// add read bytes to the output buffer
		if (readcount) {
			Answer.append(tmp2);
			count+=readcount;
			  
			// if we have more time to read, check if we got already the expected string
			if ((timeoutcounter<timeout) && (found==0)) {
				    
				// check if it's the expected answer
				if ((Answer.contains("OK\r")|| Answer.contains("ERROR")) && expect.isNull()) {
					found=1;
				}
				    
				if (!expect.isNull() && !expect.isEmpty()) {
					if (Answer.contains(expect)) {
						theLog->debug(QString("Got expected")); // + answer);
						found=1;
					}
				}
				    
				if (found) {
					//timeoutcounter=timeout-1; why so ?
					break;
				}
			}
		}
	} while (timeoutcounter<timeout); // repeat until timout

	Answ = Answer.stripWhiteSpace();
	theLog->debug(QString("R: ") + Answ);

	return 0;
}

int BtComm::getClip(QString& CallerID) 
{
	int   tryAnyway = 0;

	theLog->debug("getClip()");

	QString answer = QString::null;

	// Some phones uses AT+CLCC?, some just AT+CLCC (Sagem, Siemens-S55, SE-K750)

	if (useQuestionInCLCC == 0) {
		theLog->debug("getClip() send AT+CLCC");

		sendCommand(QString(DEF_AT_CLCC) + "\r", answer, QString::null);
		int ptr = answer.find(QString(DEF_CLCC));

		if (ptr >= 0) {
			CallerID = parse_CLCC(answer.right(answer.length() - ptr));
			return EXIT_EXACT;
		} else if (answer.find(QString("OK")) >= 0) {
			return EXIT_OK;
		} else if (answer.find(QString("ERROR")) >= 0) {
			theLog->debug("Got ERROR on AT+CLCC. Will try AT+CLCC?");
			useQuestionInCLCC = 1;      // Ask phone again; but will never ask AT+CLCC more
		} else {
			tryAnyway = 1;		// Ask phone again; to be sure
		}
	}

	if (useQuestionInCLCC == 1 || tryAnyway == 1) {
		tryAnyway = 0;

		theLog->debug("getClip() send AT+CLCC?");

		sendCommand(QString(DEF_AT_CLCC) + "?\r", answer, QString::null);
		int ptr = answer.find(QString(DEF_CLCC));

		if (ptr >= 0) {
			CallerID = parse_CLCC(answer.right(answer.length() - ptr));
			return EXIT_EXACT;
		} else if (answer.find(QString("OK")) >= 0) {
			return EXIT_OK;
		} else if (answer.find(QString("ERROR")) >= 0) {
			theLog->debug("sendClip() got ERROR");
			return EXIT_NOK;
		}
	}

	theLog->debug("sendClip() unappropriate answer: >" + answer + "<");
	return EXIT_NOK;
}

//
// Support server mode sockets
//

int BtComm::openSocketPort(int type, int port) 
{
	struct sockaddr_in tcp_addr;
	struct sockaddr_rc bt_addr;
	struct sockaddr*   socketaddr = NULL;

	style = type;

	int addFamily = 0;
	int proto     = 0;

	if (sportfd) { 
		theLog->debug("socket was already opened");
		//close(sportfd);
		//sportfd = 0;
		return 1;
	}

	if (type == SERVER_TCP) {
		addFamily = AF_INET;
		proto	  = IPPROTO_TCP;
	} else if (type == SERVER_BT) {
		addFamily = PF_BLUETOOTH;
		proto	  = BTPROTO_RFCOMM;
	} else {
		theLog->error("incorrect input");
		return -1;
	}

	if ((sportfd = socket(addFamily, SOCK_STREAM, proto)) < 0) {
		theLog->error("opening socket");
		return -1;
	}

	// Set non-blocking mode
	if (-1 == (oldflags = fcntl(sportfd, F_GETFL, 0))) {
		oldflags = 0;
	}
	fcntl(sportfd, F_SETFL, oldflags | O_NONBLOCK);

	memset((void *) &tcp_addr, 0, sizeof(tcp_addr));

	if (type == SERVER_TCP) {
		tcp_addr.sin_family = AF_INET;
		tcp_addr.sin_addr.s_addr = INADDR_ANY;
		tcp_addr.sin_port = htons(port);
	
		socketaddr=(struct sockaddr *)&tcp_addr;
	} else {	// SERVER_BT
		// bind socket to the specified port of the first available local bluetooth adapter
		bt_addr.rc_family = AF_BLUETOOTH;
		bt_addr.rc_bdaddr = *BDADDR_ANY; // bdaddr;	    // what is this ?
		bt_addr.rc_channel = (uint8_t) port;

		char tmpstring[50];
		sprintf(tmpstring, "sdptool add --channel=%i SP", port);
		system(tmpstring);
		theLog->info("registered SP for channel " + QString().setNum(port));

		socketaddr=(struct sockaddr *)&bt_addr;
	}

	if (bind(sportfd, (struct sockaddr *) socketaddr, sizeof(*socketaddr)) < 0) {
		theLog->error("on binding");
		printf("ERROR: on binding %d->%s\n", errno, strerror(errno));
	
		return -1;
	}

	return 1;
}

int BtComm::listenAndAcceptSocketConn(int type) 
{
	int cnt; //, flags;
	socklen_t sz;
	struct sockaddr_in ip_addr;
	struct sockaddr_rc bt_addr;
	bdaddr_t ba;

	//theLog->info("listenAndAcceptSocketConn");
	cnt = 0;

	struct sockaddr* socketaddr=(struct sockaddr *)&bt_addr;
	sz = sizeof(bt_addr);
	if (type == SERVER_TCP) {
		socketaddr=(struct sockaddr *)&ip_addr;
		sz = sizeof(ip_addr);
	}

	listen(sportfd,5);
	 
	portfd = accept(sportfd, (struct sockaddr *) &socketaddr, &sz);
		 
	if (portfd == -1 && errno == EAGAIN) {  	  
		return -1;
	}	 
	 
	if (portfd < 0) {
		theLog->error("on accept");
		printf("ERROR: on accept %d\n", errno);
		return -1;
	/*} else {
		// Set non-blocking mode
		if (-1 == (flags = fcntl(portfd, F_GETFL, 0))) {
			flags = 0;
		}
		fcntl(portfd, F_SETFL, flags | O_NONBLOCK);
	*/
	}
	theLog->info("listenAndAcceptSocketConn: accepted");

	Connected = true;

	if (type == SERVER_BT) {
		baswap(&ba, &bt_addr.rc_bdaddr);
		theLog->info(QString("listenAndAcceptSocketConn: remote BT address is ") + batostr(&ba));
	}

	return 1;
}

int BtComm::writeSocketConn(QString& command) 
{
	//theLog->debug("writeSocketConn");

	// send command
	if (!(command.isNull() || command.isEmpty())) {
		theLog->debug("writeSocketConn " + command);

		int n = write(portfd, command.ascii(), strlen(command.ascii()));

		if (n < 0) {
			theLog->error("error writing to socket");
			return EXIT_NOK;
		}
	}
	return EXIT_OK;
}

int  BtComm::writeBytesSocketConn(QString& command) 
{
	theLog->debug("writeBytesSocketConn");

	// send command
	if (!(command.isNull() || command.isEmpty())) {
		theLog->debug("writeBytesSocketConn " + command);

		int section = 0;
		QString s = command.section(",", section, section).simplifyWhiteSpace();
		while (!s.isNull()) {
			
			//theLog->debug("Next byte is >" + s + "<");
		
			unsigned char byte2write[2];
			byte2write[0] = (unsigned char) atoi(s.ascii());
			byte2write[1] = '\0';
		
			if (write(portfd, byte2write, 1) < 0) {
				theLog->error("error writing to socket");
				return EXIT_NOK;
			}
			section++;
			s = command.section( ",", section, section).simplifyWhiteSpace();
		}
	}
	return EXIT_OK;
}

QString BtComm::parse_CLCC(QString clcc)
{
	// There are could be more than one call at time ... but here we will get then one-by-one
	// +CLCC: 1,,4,0,0,"<phone number here>",129,
	QString ID;
	QString ptr = clcc.section(',',5,5);

	if (ptr.isNull() || ptr.isEmpty()) {	// Call was finished
		return QString("FINISHED");
	}

	if (ptr.at(0) == QChar('"')) {
		ID = ptr.mid(1, ptr.length() - 2);
	} else {
		ID = ptr;
	}

	if(ID.isNull() || ID.isEmpty()) {	// Empty CallerID
		theLog->debug("parse_CLCC >NULL callerID<");
		return QString("NO CALLER ID");
	} else {
		theLog->debug(QString("parse_CLCC >") + ID + "<");
	}

	return ID;
}

/*int BtComm::readSocketConn(QString& bf) 
{
	int n;

	//theLog->debug("readSocketConn");
	char buffer[256];

	bzero(buffer,256);
	n = readTmo(buffer, 255, 100000); 
	if (n == 0) return EOF - 1;

	if (n == EOF) {
		// Connect closed
		theLog->info("readSocketConn: connection closed");
		Connected = false;
		return EOF;
	}


	if (n > 0) {
		bf = buffer;
		theLog->debug("Got " + bf);
	}
	return n;
}
*/
